//
//	LINKFM.CPP
//	Link Folder to Message
//
//	Given a folder path, creates a .XNK attachment which can be used to
//	launch an Exchange viewer on that folder, or attached to a
//	message as a folder link. This is the same type of file as is
//	generated by the File.Create Shortcut command in Exchange.
//
//	Command line arguments are as follows:
char szUsage[] = 
	"Usage: linkfm folder-path\r\n"
	"Examples:\r\n"
	"    linkfm Public Folders\\Fishing Spots\r\n"
	"    linkfm Inbox\r\n"
	"Note: don't try to use quotes to preserve spaces.\r\n";
char szApp[] = "Link Folder to Message";
//
//	Exchange will create a new window to view the folder if you 
//	double-click the message, then double-click on the attached .XNK
//	file.
//
//	The following information on links to messages and folders is not part
//	of the MAPI specification. It applies only to the Microsoft Exchange
//	client version 4.0. Other MAPI clients, as well as future versions
//	of the Microsoft Exchange client, may not be compatible with links
//	built according to the information supplied here.
//

//	Format string for test message body.
char szTextFmt[]	= "Link to folder '%.256s'\r\n \r\n";
char szSubjectFmt[]	= "Link to '%.40s'";
char szFileFmt[]	= "%.27s.XNK";

#include <windows.h>
#include <mapix.h>
#include <mapiutil.h>
#include "_links.h"
#include "_resrc.h"

STDAPI	LinkFolderToNewMessage(LPMESSAGE pmsg, LPENTRYLIST pel,
		LPSTR szFileName, LPSTR szFolder, HICON hicon, ULONG ulPos);

int CALLBACK
WinMain (HINSTANCE hInstance, HINSTANCE hPrevInst, LPSTR lpszCmd, int nCmdShow)
{
	HRESULT			hr;
	BOOL			fMAPIInit = FALSE;
	LPMAPISESSION	pses = NULL;
	LPSTR			szPath = NULL;
	LPSTR			sz;
	LPMDB			pmdbDefault = NULL;
	LPMAPIFOLDER	pfldInbox = NULL;
	LPMESSAGE		pmsg = NULL;
	LPENTRYLIST		pel = NULL;
	LPSPropValue	pval = NULL;
	LPSPropProblemArray pprobs = NULL;
	SPropValue		rgval[4];
	SYSTEMTIME		st;
	FILETIME		ft;
	HICON			hicon = 0;
	ULONG			ichRender;
	char			szMsgText[512];
	char			szMsgSubject[64];
	char			szFile[256];
	LPSTR			szFolder;
	
	//	Skip to the first argument.
	for (sz = lpszCmd; *sz && *sz == ' '; ++sz)
		;
	if (!*sz)
		goto usage;
	//	Copy the path because we're going to munge it.
	if (!(szPath = strdup(sz)))
		goto ret;
	//	Find the last node for display purposes.
	for (sz = szFolder = lpszCmd; sz = strpbrk(szFolder, szPathSep);
		sz = szFolder)
	{
		szFolder = sz + 1;
	}
	wsprintf(szFile, szFileFmt, szFolder);

	//	Start up MAPI, log on, and open the Inbox.
	if (hr = MAPIInitialize(NULL))
		goto ret;
	fMAPIInit = TRUE;
	if (hr = MAPILogonEx(0, NULL, NULL, MAPI_USE_DEFAULT | MAPI_NO_MAIL,
		&pses))
		goto ret;
	if (hr = OpenInbox(pses, &pmdbDefault, &pfldInbox))
		goto ret;

	//	Crack the folder path.
	if (hr = TraceFolderPath(pses, szPath, &pel, NULL))
		goto ret;

	//	Prepare certain cosmetic features
	hicon = LoadIcon(hInstance, MAKEINTRESOURCE(rsidFolderLink));
	wsprintf(szMsgText, szTextFmt, szFolder);
	{
		LPSTR	szT;

		szT = strrchr(szMsgText, ' ');
		ichRender = szT ? szT - szMsgText : 0xffffffff;
	}
	wsprintf(szMsgSubject, szSubjectFmt, szFolder);

	//	Create a new message in the Inbox, with recognizable subject
	//	and text.
	if (hr = pfldInbox->CreateMessage(NULL, MAPI_DEFERRED_ERRORS, &pmsg))
		goto ret;
	rgval[0].ulPropTag = PR_BODY;
	rgval[0].Value.lpszA = szMsgText;
	rgval[1].ulPropTag = PR_SUBJECT;
	rgval[1].Value.lpszA = szMsgSubject;
	rgval[2].ulPropTag = PR_MESSAGE_DELIVERY_TIME;
	GetSystemTime(&st);
	SystemTimeToFileTime(&st, &ft);
	rgval[2].Value.ft = ft;
	rgval[3].ulPropTag = PR_MESSAGE_CLASS;
	rgval[3].Value.lpszA = "IPM.Note";
	if (hr = pmsg->SetProps(4, rgval, &pprobs))
		goto ret;
	if (pprobs)
	{
		hr = ResultFromScode(pprobs->aProblem[0].scode);
		goto ret;
	}

	//	Lay down the folder link.
	if (hr = LinkFolderToNewMessage(pmsg, pel, szFile, szFolder, 
			hicon, ichRender))
		goto ret;

	//	Save the message (don't send it).
	if (hr = pmsg->SaveChanges(0))
		goto ret;

ret:
	//	Clean up after ourselves.
	MAPIFreeBuffer(pel);
	MAPIFreeBuffer(pval);
	MAPIFreeBuffer(pprobs);
	UlRelease(pmsg);
	UlRelease(pfldInbox);
	UlRelease(pmdbDefault);
	UlRelease(pses);
	if (fMAPIInit)
		MAPIUninitialize();
	if (szPath)
		free(szPath);
	if (hicon)
		DestroyIcon(hicon);
	
	//	Do the usual exit status thing: 
	//	0 => OK, nonzero => failure
	return ResultFromScode(hr) != S_OK;

usage:
	MessageBox(NULL, szUsage, szApp, MB_OK | MB_ICONSTOP);
	goto ret;
}

//
//	LinkFolderToNewMessage inserts a link to the folder whose
//	path is given by pel into a new message in the default Inbox.
//	The strings szFolderPath and szFileName are used for display,
//	but szFileName must have the extension ".XNK" in order for Exchange
//	to launch the attachment properly.
//
//	If you have an icon for the folder link, pass its handle in hicon.
//
STDAPI
LinkFolderToNewMessage(LPMESSAGE pmsg, LPENTRYLIST pel, LPSTR szFileName,
	LPSTR szFolder, HICON hicon, ULONG ulPos)
{
	HRESULT			hr;
	LPATTACH		patch = NULL;
	LPSTREAM		pstream = NULL;
	LPSPropProblemArray pprobs = NULL;
	SPropValue		rgval[5];
	char			szTitle[300];
	ULONG			ul;

	//	Create a new attachment in the message.
	if (hr = pmsg->CreateAttach(NULL, MAPI_DEFERRED_ERRORS, &ul, &patch))
		goto ret;

	//	Store the folder link in the attachment's content property.
	if (hr = patch->OpenProperty(PR_ATTACH_DATA_BIN, &IID_IStream,
		STGM_DIRECT | STGM_WRITE | STGM_SHARE_EXCLUSIVE | STGM_CREATE,
			MAPI_CREATE | MAPI_MODIFY, (LPUNKNOWN FAR *) &pstream))
		goto ret;
	if (hr = WriteFolderLink(pel, pstream))
		goto ret;
	if (hr = pstream->Commit(0))
		goto ret;
	pstream->Release();
	pstream = NULL;	

	//	Make name strings for the attachment title, message subject,
	//	and message text. There is nothing magic about these strings.
	//	The code that sets PR_RENDERING_POSITION for the attachment
	//	expects that there is a blank at the end of szText, which
	//	some mail clients will replace with the attachment's icon
	//	when displaying the message.
	wsprintf(szTitle, "Link to %s", szFolder);

	//	Set up the display name and other attachment properties.
	//	Some notes:
	//	1. The attachment method is BY_VALUE even though we're making
	//	a folder link. To MAPI, this is just an attached file with
	//	some data; Exchange interprets it as a link.
	//	2. Some mail clients (especially on 16-bit Windows) must have
	//	an 8.3 filename in PR_ATTACH_FILENAME. The file names must
	//	have the extension .XNK in order for Exchange to launch.
	//	3. The value of PR_RENDERING_POSITION must be less than the
	//	length of the message's PR_BODY; we will write szText
	//	there in a little bit.
	rgval[0].ulPropTag = PR_ATTACH_METHOD;
	rgval[0].Value.l = ATTACH_BY_VALUE;
	rgval[1].ulPropTag = PR_DISPLAY_NAME;
	rgval[1].Value.lpszA = szTitle;
	rgval[2].ulPropTag = PR_ATTACH_LONG_FILENAME;
	rgval[2].Value.lpszA = szFileName;
	//$	it's not a real filename - how to get good long/short names?
	rgval[3].ulPropTag = PR_ATTACH_FILENAME;
	rgval[3].Value.lpszA = szFileName;
	rgval[4].ulPropTag = PR_RENDERING_POSITION;
	rgval[4].Value.l = ulPos;
	if (hr = patch->SetProps(5, rgval, &pprobs))
		goto ret;
	if (pprobs)
	{
		//	Any problem is fatal. Use the first error code we're given.
		hr = ResultFromScode(pprobs->aProblem[0].scode);
		goto ret;
	}

	//	Store the attachment icon.
	if (hicon)
	{
		if (hr = SetAttachmentIcon(hicon, patch))
			goto ret;
	}

	//	Commit the attachment.
	if (hr = patch->SaveChanges(0))
		goto ret;
	patch->Release();
	patch = NULL;

ret:
	//	Clean up. The order of the Release() calls is important.
	UlRelease(pstream);
	UlRelease(patch);

	MAPIFreeBuffer(pprobs);

	return hr;
}

